#include "sif.h"
#include <stdio.h>

const int hash_name(const char* name, int len)
{
   int i = 0;
   while(len) /* make a 32 bit hash */
       i ^= (*name++ << ((len--*7) % (25)));
   return i;
}

/* this prints message then exits forth */
int bye_message(sif* s, const char* message)
{
  fprintf(stdout, "\n%s\n", message);
  bye(s);
}

/* this just exits forth */
int bye(sif* s)
{
  free(s->mem);
  exit(1);
}

void standard_flush(sif* s)
{
  fflush(stdout);
}

int standard_input(sif* s, void* buffer, int max) /* returns true for success (0 means end of input) */
{
  int c = getchar();

  if(c == EOF)
    return 0;
  
  *(char*)buffer = (char)c;
  return 1;
  /* return fgets(buffer, max, stdin) > 0; */
}

void standard_output(sif* s, void* buffer, int count) /* only returns on success */
{
  if(fwrite(buffer, 1, count, stdout) != count)
    {
      bye_message(s, "Write failed"); /* fix */
      exit(77);
    }
  fflush(stdout);
}

/* initialize s with, memory block of size bytes. set all members to default values */
int sif_init(sif* s, void* mem, int size)
{
  cell c;

  s->mem = mem;

  switch(CELL_SIZE)
    {
    case 1:
      s->character_mask = 0; /* in 8 bit (if it even works) the memory block will be 256 bytes */
      break;
    case 2:
      s->character_mask = ~(size | (size<<1) | (size<<2) | (size<<3) | (size<<4) | (size<<5) | (size<<6));
      break;
    case 4:
      c = size | (size<<1) | (size<<2) | (size<<3) | (size<<4) | (size<<5);
      s->character_mask = ~(c | (c<<6) | (c<<12) | (c<<18));
    }

  s->cell_mask = (CELL_SIZE - 1) ^ s->character_mask;


  s->stack      = (cell*)s->mem;
  s->rstack     = s->stack + STACK_SIZE;
  s->word       = (character*)(s->rstack + RSTACK_SIZE);
  s->heap       = s->word + MAX_WORD_SIZE;
  s->dictionary = s->mem + size;

  s->cur_stack      = 0; s->cur_stack--;
  s->cur_rstack     = 0;
  s->cur_heap       = (int)(s->heap) - (int)(s->mem);
  s->cur_dictionary = 0;

  s->code_position  = s->cur_heap; /* does this matter? */

  s->input  = standard_input;
  s->output = standard_output;
  s->flush = standard_flush;

  s->state  = STATE_INTERPRET;
  s->heap_dict_size = size - CELL_SIZE * (STACK_SIZE + RSTACK_SIZE);

  init_dictionary(s);

  return 0;
}
